import argparse
from pathlib import Path

from .migrate import migrate


def main():
    parser = init_argparse()
    args = parser.parse_args()
    source = args.source
    target = args.target if args.target is not None else args.source
    migrate(source, target)


def init_argparse() -> argparse.ArgumentParser:
    arg_parser = argparse.ArgumentParser(
        prog="Framebot migration tool",
        usage=f"migrate -s source -t target...",
        description="Migrates the local file created from old framebot versions for use with the newer ones."
    )
    arg_parser.add_argument("-s", "--source", type=Path, metavar="Source directory",
                            help="Directory where the legacy framebot files are stored.", required=True)
    arg_parser.add_argument("-t", "--target", type=Path, metavar="Target directory",
                            help="Directory where the new framebot files are to be stored.", required=False)
    return arg_parser


if __name__ == "__main__":
    main()



import logging
import shutil
import unittest
import tempfile
from pathlib import Path

import jsonpickle

from framebot.utils import get_logger, LoggingObject, safe_json_dump


class TestUtils(unittest.TestCase):

    def setUp(self) -> None:
        self.test_dir = Path(tempfile.mkdtemp())

    def tearDown(self) -> None:
        shutil.rmtree(self.test_dir)

    def test_safe_json_dump(self):
        test_json_file = self.test_dir.joinpath("test_json.json")
        test_obj = {"intParam": 1, "strParam": "test"}
        safe_json_dump(test_json_file, test_obj)
        self.assertTrue(test_json_file.exists())
        with open(test_json_file, "r") as f:
            read_obj = f.read()
            read_obj = jsonpickle.decode(read_obj)
            self.assertEqual(read_obj, test_obj)
        test_json_file.unlink()
        safe_json_dump(str(test_json_file), test_obj)
        self.assertTrue(test_json_file.exists())
        with open(test_json_file, "r") as f:
            read_obj = f.read()
            read_obj = jsonpickle.decode(read_obj)
            self.assertEqual(read_obj, test_obj)

    def test_get_logger(self):
        logger_name = "test_logger"
        level = logging.DEBUG
        logger = get_logger(logger_name, level)
        self.assertEqual(logger.name, logger_name)
        self.assertEqual(logger.level, level)

        # default level
        logger = get_logger(logger_name)
        self.assertEqual(logger.level, logging.INFO)
        # None parameters
        with self.assertRaises(ValueError):
            get_logger(None, level)
        with self.assertRaises(ValueError):
            get_logger(logger_name, None)

    def test_logging_object(self):
        logging_object = LoggingObject()
        self.assertEqual(logging_object.logger.name, type(logging_object).__name__)

        class LoggingObjectSubclass(LoggingObject):
            pass

        logging_object = LoggingObjectSubclass()
        self.assertEqual(logging_object.logger.name, LoggingObjectSubclass.__name__)


if __name__ == '__main__':
    unittest.main()



import logging
import shutil
import sys
from logging import Logger
from pathlib import Path
from typing import Union, Any

import jsonpickle


def safe_json_dump(fpath: Union[str, Path], obj: Any) -> None:
    """
    Utility function used to avoid json file corruption in case of abrupt termination of the script
    :param fpath: path where the json has to be saved
    :param obj: the content to be saved
    """
    if issubclass(type(fpath), Path):
        fpath = str(fpath)
    safe_path = fpath + "_safe"
    with open(safe_path, "w") as f:
        json_str = jsonpickle.dumps(obj, indent=4)
        f.write(json_str)
    shutil.move(safe_path, fpath)


def load_obj_from_json_file(fpath: Union[str, Path]) -> Any:
    """

    :param fpath:
    :return:
    """
    with open(fpath, "r") as f:
        str_result = f.read()
        result = jsonpickle.decode(str_result)
    return result


def get_logger(name: str, level: int = logging.INFO) -> Logger:
    """

    :param name:
    :param level:
    :return:
    """
    if name is None or level is None:
        raise ValueError("name and level must not be None!")
    logger = Logger(name)
    logger.setLevel(level)
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('[%(asctime)s|%(name)s|%(levelname)s] %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


class LoggingObject(object):
    """
    Object that comes equipped with a logger
    """
    logger: Logger

    def __init__(self):
        self.logger = get_logger(type(self).__name__)



import argparse
import configparser
import os
import platform
import shutil
import sys
from datetime import timedelta
from importlib import resources
from pathlib import Path
from typing import List

from .framebots import SimpleFrameBot
from .plugins import BestOfReposter, MirroredFramePoster, FrameBotPlugin, AlternateFrameCommentPoster
from .social import FacebookHelper
from .utils import get_logger

logger = get_logger("main")


def _init_best_of_reposter(config: configparser.ConfigParser, facebook_helper: FacebookHelper,
                           movie_title: str, working_dir: Path, plugins: List[FrameBotPlugin]) -> None:
    if config.has_section("best_of_album_uploader"):
        bof_uploader_settings = config["best_of_album_uploader"]
        best_of_reposting_enabled = bof_uploader_settings.getboolean("enabled")
        if best_of_reposting_enabled:
            reactions_threshold = bof_uploader_settings.getint("reactions_threshold")
            wait_hours = bof_uploader_settings.getint("wait_hours")
            best_of_album_id = bof_uploader_settings.get("best_of_album_id")
            plugins.append(BestOfReposter(
                album_id=best_of_album_id, facebook_helper=facebook_helper, video_title=movie_title,
                reactions_threshold=reactions_threshold, time_threshold=timedelta(hours=wait_hours),
                working_dir=working_dir)
            )


def _init_mirrored_frame_poster(config: configparser.ConfigParser, facebook_helper: FacebookHelper,
                                bot_name: str, plugins: List[FrameBotPlugin]) -> None:
    if config.has_section("mirroring"):
        mirroring_settings = config["mirroring"]
        mirroring_enabled = mirroring_settings.getboolean("enabled")
        if mirroring_enabled:
            mirroring_ratio = mirroring_settings.getfloat("ratio")
            mirror_album_id = mirroring_settings["mirror_album_id"]
            plugins.append(
                MirroredFramePoster(album_id=mirror_album_id, facebook_helper=facebook_helper,
                                    bot_name=bot_name, ratio=mirroring_ratio)
            )


def _init_config_parser(config_directory: Path) -> configparser.ConfigParser:
    config_path = config_directory.joinpath("config.ini")

    if not config_path.exists():
        logger.warning(f"No 'config.ini' file found in the provided directory: {config_directory}.")
        logger.info("Would you like to generate a config.ini file in the provided location? (yes/no)")
        user_input = input()

        yes_choices = ['yes', 'y']

        if user_input.lower() in yes_choices:
            logger.info(f"Generating config.ini file in {config_directory}...")
            shutil.copy(resources.files("framebot.resources").joinpath("config.ini"), config_directory)
            logger.info("config.ini file generated. Please edit the configuration values according to your preferences"
                        " and then press any key.")
            input()
        else:
            logger.warning("The bot needs a configuration file in order to initialize. Please create or generate one in"
                           " your directory of choice.")
            exit()

    config = configparser.ConfigParser()
    config.read(config_path, encoding='utf-8')
    return config


def _configure_window(movie_title: str) -> None:
    op_sys = platform.system()
    window_title = f"Framebot - {movie_title}"
    if op_sys == "Windows":
        os.system(f"title {window_title}")
    elif op_sys == "Linux":
        sys.stdout.write(f"\x1b]2;{window_title}\x07")


def _init_facebook_helper(config: configparser.ConfigParser) -> FacebookHelper:
    facebook_settings = config["facebook"]
    page_id = facebook_settings.get("page_id")
    access_token = facebook_settings.get("access_token")
    return FacebookHelper(access_token=access_token, page_id=page_id)


def _init_alternate_frame_poster(config: configparser.ConfigParser, facebook_helper: FacebookHelper,
                                 delete_files: bool, working_dir: Path, plugins: List[FrameBotPlugin], ) -> None:
    if config.has_section("alternate_frame_poster"):
        alternate_frame_poster_settings = config["alternate_frame_poster"]
        alternate_enabled = alternate_frame_poster_settings.getboolean("enabled")
        if alternate_enabled:
            alternate_directory = Path(alternate_frame_poster_settings.get("alternate_frames_directory"))
            if not alternate_directory.is_absolute():
                alternate_directory = working_dir.joinpath(alternate_directory).resolve()
            comment_text = alternate_frame_poster_settings.get("comment_text")
            plugins.append(
                AlternateFrameCommentPoster(alternate_frames_directory=alternate_directory,
                                            facebook_helper=facebook_helper, delete_files=delete_files,
                                            message_generator=comment_text)
            )


def main():
    parser = init_argparse()
    args = parser.parse_args()

    config_directory = args.directory

    config = _init_config_parser(config_directory)

    bot_settings = config["bot_settings"]
    upload_interval = timedelta(seconds=bot_settings.getint("upload_interval"))
    frames_directory = Path(bot_settings.get("frames_directory"))
    frames_ext = bot_settings.get("frames_ext")
    frames_naming = bot_settings.get("frames_naming")
    movie_title = bot_settings.get("movie_title")
    bot_name = bot_settings.get("bot_name")
    delete_files = bot_settings.getboolean("delete_files")

    _configure_window(movie_title)

    working_dir = Path(config_directory)
    if not frames_directory.is_absolute():
        frames_directory = working_dir.joinpath(frames_directory).resolve()
    facebook_helper = _init_facebook_helper(config)

    plugins = []

    _init_best_of_reposter(config, facebook_helper, movie_title, working_dir, plugins)
    _init_mirrored_frame_poster(config, facebook_helper, bot_name, plugins)
    _init_alternate_frame_poster(config, facebook_helper, delete_files, working_dir, plugins)

    bot = SimpleFrameBot(facebook_helper=facebook_helper, video_title=movie_title,
                         upload_interval=upload_interval, delete_files=delete_files, bot_name=bot_name,
                         frames_ext=frames_ext,
                         frames_directory=frames_directory, frames_naming=frames_naming, plugins=plugins,
                         working_dir=working_dir)
    bot.start()


def init_argparse() -> argparse.ArgumentParser:
    arg_parser = argparse.ArgumentParser(
        prog="Framebot",
        usage=f"framebot -d directory",
        description="Starts a framebot using configuration found inside a 'config.ini' present in the provided"
                    " directory."
    )
    arg_parser.add_argument("-d", "--directory", type=Path, metavar="Configuration directory",
                            help="Where the 'config.ini' file is stored.", required=True)
    return arg_parser


if __name__ == '__main__':
    main()



import datetime
import re
import shutil
import sys
import unittest
from datetime import timedelta, datetime
from distutils.dir_util import copy_tree
from pathlib import Path
from unittest.mock import Mock, patch, mock_open

from framebot.framebots import SimpleFrameBot, _get_filename
from framebot.model import FacebookFrame
from framebot.plugins import FrameBotPlugin
from framebot.social import FacebookHelper
from test.utils_for_tests import FileWritingTestCase

from test import RESOURCES_DIR


class TestStaticMethods(unittest.TestCase):

    def test_get_filename(self):
        filename = "dummy"
        extension = ".jpg"
        full_filename = f"{filename}{extension}"
        self.assertEqual(filename, _get_filename(filename))
        self.assertEqual(full_filename, _get_filename(full_filename))
        self.assertEqual(filename, _get_filename(
            Path("root").joinpath("exampledir").joinpath(filename)))


class TestSimpleFrameBot(FileWritingTestCase):

    def setUp(self) -> None:
        super().setUp()
        self.mock_helper = Mock(spec=FacebookHelper)
        self.video_title = "Test video title"
        frames_directory = self.test_dir.joinpath("frames")
        frames_directory.mkdir(parents=True, exist_ok=True)
        self.testee = SimpleFrameBot(
            facebook_helper=self.mock_helper,
            video_title=self.video_title,
            working_dir=self.test_dir
        )
        self.mock_plugin = Mock(spec=FrameBotPlugin)

    def _copy_frames_directory(self):
        if sys.version_info[0] == 3 and sys.version_info[1] == 7:
            copy_tree(str(RESOURCES_DIR.joinpath("framebots").joinpath("simple_framebot").joinpath("frames")),
                      str(self.test_dir.joinpath("frames")))
        else:
            shutil.copytree(RESOURCES_DIR.joinpath("framebots").joinpath("simple_framebot").joinpath("frames"),
                            self.test_dir.joinpath("frames"), dirs_exist_ok=True)

    def test_init(self):
        self.assertEqual([], self.testee.plugins)

    def test_frames_naming(self):
        with self.assertRaises(ValueError):
            self.testee.frames_naming = "nonsense"

        self.testee.frames_ext = "jpg"
        self.testee.frames_naming = "$N$"
        self.assertEqual(re.compile("^(\\d+)\\.jpg$"), self.testee.frames_naming)

        self.testee.frames_naming = "My cool movie frame $N$ of 123543"
        self.assertEqual(re.compile("^My cool movie frame (\\d+) of 123543\\.jpg$"), self.testee.frames_naming)
        self.assertIsNotNone(self.testee.frames_naming.match("My cool movie frame 000010 of 123543.jpg"))

    def test_init_frames(self):
        # empty frame list
        self.testee._init_frames()
        self.assertEqual(0, len(self.testee.frames))

        # frame list with some frames
        self._copy_frames_directory()
        self.testee.last_frame_uploaded = -1
        self.testee._init_frames()

        expected_frames_paths = list(filter(lambda p: bool(self.testee._frames_naming.match(p.name)),
                                            self.test_dir.joinpath("frames").glob("*.jpg")))
        expected_frames_paths.sort(key=lambda p: int(re.search("^(\\d+)\\.jpg$", p.name).group(1)))
        self.assertEqual(len(expected_frames_paths), len(self.testee.frames))
        for i, frame in enumerate(self.testee.frames):
            self.assertEqual(i + 1, frame.number)
            self.assertEqual(expected_frames_paths[i], frame.local_file)

        # start from later frame
        self.testee.last_frame_uploaded = 5
        self.testee._init_frames()

        self.assertEqual(len(expected_frames_paths) - self.testee.last_frame_uploaded, len(self.testee.frames))
        for i, frame in enumerate(self.testee.frames):
            self.assertEqual(i + self.testee.last_frame_uploaded + 1, frame.number)
            self.assertEqual(expected_frames_paths[i + self.testee.last_frame_uploaded], frame.local_file)

    @patch("builtins.open", new_callable=mock_open, read_data="13")
    def test_init_status(self, mock_file: Mock):
        # no last frame uploaded file
        self.testee.last_frame_uploaded_file = Mock(spec=Path)
        mock_exists: Mock = self.testee.last_frame_uploaded_file.exists
        mock_exists.return_value = False
        self.testee._init_status()
        self.assertEqual(-1, self.testee.last_frame_uploaded)
        mock_file.assert_not_called()
        mock_exists.assert_called_once_with()

        # file found
        mock_exists.reset_mock()
        mock_exists.return_value = True
        self.testee._init_status()
        self.assertEqual(13, self.testee.last_frame_uploaded)
        mock_file.assert_called_once_with(self.testee.last_frame_uploaded_file)
        mock_exists.assert_called_once_with()

    def test_get_frame_index_number(self):
        self.assertEqual(1, self.testee._get_frame_index_number("1.jpg"))
        self.assertEqual(1, self.testee._get_frame_index_number("001.jpg"))
        self.assertEqual(123, self.testee._get_frame_index_number(Path("example").joinpath("123.jpg")))
        self.assertEqual(123, self.testee._get_frame_index_number(Path("example").joinpath("123.jpg").absolute()))

        complicated_name_string = "I like complicated frame names $N$. Frame number was right before this sentence!"
        self.testee.frames_naming = complicated_name_string

        self.assertEqual(2, self.testee._get_frame_index_number(
            Path("example").joinpath(complicated_name_string.replace("$N$", "0002") + ".jpg")))

    def test_get_default_message(self):
        total_frames_number = 20
        frame_number = 12

        self.testee.total_frames_number = total_frames_number
        message = self.testee._get_default_message(frame_number)
        self.assertEqual(f"{self.video_title}\nFrame {frame_number} of {total_frames_number}", message)

    def test_update_last_frame_uploaded(self):
        test_last_frame_uploaded = 10
        self.testee._update_last_frame_uploaded(test_last_frame_uploaded)

        self.assertEqual(test_last_frame_uploaded, self.testee.last_frame_uploaded)
        test_last_frame_uploaded_path = self.test_dir.joinpath(self.testee.last_frame_uploaded_file)
        self.assertTrue(test_last_frame_uploaded_path.exists())
        with open(test_last_frame_uploaded_path) as f:
            self.assertEqual(str(test_last_frame_uploaded), f.read())

    def test_start(self):
        self.testee.plugins.append(self.mock_plugin)
        self.testee._upload_loop = Mock()
        self.testee.last_frame_uploaded_file = Mock(spec=Path)

        self.testee.start()

        self.testee._upload_loop.assert_called_once()
        self.testee.last_frame_uploaded_file.unlink.assert_called_once_with()
        self.mock_plugin.before_upload_loop.assert_called_once()
        self.mock_plugin.after_upload_loop.assert_called_once()

    @patch("os.remove")
    def test_upload_loop(self, mock_remove: Mock):
        self.testee.plugins.append(self.mock_plugin)
        self.testee._upload_frame = Mock(spec=self.testee._upload_frame)
        self.testee._determine_adjusted_pause = Mock(spec=self.testee._determine_adjusted_pause)
        self.testee._determine_adjusted_pause.return_value = timedelta(0)
        self.testee.upload_interval = timedelta(microseconds=0)
        self._copy_frames_directory()
        self.testee._init_frames()
        frames_number = self.testee.total_frames_number
        self.assertGreater(frames_number, 0)

        self.testee._upload_loop()
        self.assertEqual(frames_number, self.testee._upload_frame.call_count)
        self.assertEqual(frames_number, self.mock_plugin.before_frame_upload.call_count)
        self.assertEqual(frames_number, self.mock_plugin.after_frame_upload.call_count)
        mock_remove.assert_not_called()

        self.testee.delete_files = True
        self.testee.last_frame_uploaded = -1
        self.testee._init_frames()
        self.testee._upload_loop()
        self.assertEqual(frames_number, mock_remove.call_count)

    def test_upload_frame(self):
        test_frame = FacebookFrame(number=1, local_file="dummy.jpg")
        self.testee._upload_frame(test_frame)

        self.assertEqual(self.testee._get_default_message(test_frame.number), test_frame.text)
        self.assertIsNotNone(test_frame.photo_id)
        self.assertIsNotNone(test_frame.url)
        self.assertEqual(test_frame.number, self.testee.last_frame_uploaded)
        self.mock_helper.post_photo.assert_called_once_with(test_frame.local_file, test_frame.text)

    @patch("framebot.framebots.datetime", spec=datetime)
    def test_determine_adjusted_pause(self, mock_datetime: Mock):
        fake_now = datetime(year=2020, month=1, day=1)
        test_frame = FacebookFrame(number=1, local_file="dummy.jpg")
        test_upload_interval = timedelta(minutes=5)
        self.testee.upload_interval = test_upload_interval
        mock_datetime.now.return_value = fake_now

        test_frame.post_time = fake_now - test_upload_interval
        self.assertEqual(timedelta(seconds=0), self.testee._determine_adjusted_pause(test_frame))

        test_frame.post_time = fake_now
        self.assertEqual(test_upload_interval, self.testee._determine_adjusted_pause(test_frame))

        test_frame.post_time = fake_now - 2 * test_upload_interval
        self.assertEqual(timedelta(seconds=0), self.testee._determine_adjusted_pause(test_frame))




if __name__ == '__main__':
    unittest.main()



"""
Contains implementations for framebot plugins
"""
from __future__ import annotations

import copy
import shutil
import time
import slugify
from datetime import timedelta, datetime
from random import random
from typing import List, Type, Dict, Union, Callable

from PIL import Image, ImageOps
from pyfacebook import FacebookError

from pathlib import Path
import os

from . import DEFAULT_WORKING_DIR
from . import utils
from .model import FacebookFrame
from .social import FacebookHelper


class FrameBotPlugin(utils.LoggingObject):
    """
    A plugin to inject custom extra behavior into a framebot. It has handles for before and after the upload loop
    and the single frame posting
    """

    def __init__(self, depends_on: List[Type[FrameBotPlugin]] = None):
        """
        Constructor
        :param depends_on: Signals this plugin depends on other plugins, and thus it cannot be used without these,
        and also must act after its dependencies. Behavior for this is yet to be implemented
        """
        super().__init__()
        self.logger.info(f"Initializing plugin {type(self).__name__}")
        if depends_on is None:
            depends_on = []
        self.depends_on: List[Type[FrameBotPlugin]] = depends_on
        self.dependencies: Dict[FrameBotPlugin] = {}

    def before_upload_loop(self) -> None:
        """
        Behavior to be executed before the upload loop starts.
        """
        self.logger.debug(f"No operation defined for 'before_upload_loop'")

    def after_upload_loop(self) -> None:
        """
        Behavior to be executed after the upload loop ends.
        """
        self.logger.debug(f"No operation defined for 'after_upload_loop'")

    def before_frame_upload(self, frame: FacebookFrame) -> None:
        """
        Behavior to be executed before a frame is uploaded
        :param frame: the frame to be uploaded
        """
        self.logger.debug(f"No operation defined for 'before_frame_upload'")

    def after_frame_upload(self, frame: FacebookFrame) -> None:
        """
        Behavior to be executed after a frame is uploaded
        :param frame: the uploaded frame
        """
        self.logger.debug(f"No operation defined for 'after_frame_upload'")


class FileWritingFrameBotPlugin(FrameBotPlugin):
    """
    A FrameBotPlugin that also needs to perform file writing operations within a working directory.
    """

    def __init__(self, depends_on: List[Type[FrameBotPlugin]] = None, working_dir: Path = None):
        """
        Constructor
        :param depends_on: Signals this plugin depends on other plugins, and thus it cannot be used without these,
        and also must act after its dependencies. Behavior for this is yet to be implemented
        :param working_dir: Local working directory for the plugin
        """
        super().__init__(depends_on=depends_on)
        if working_dir is None:
            working_dir = DEFAULT_WORKING_DIR
        self.working_dir: Path = working_dir.joinpath("plugins").joinpath(type(self).__name__).resolve(strict=False)
        os.makedirs(self.working_dir, exist_ok=True)


class BestOfReposter(FileWritingFrameBotPlugin):
    """
    Reposts frames that had a reaction count over a certain threshold, after a defined time threshold from the first
    post.
    """

    def __init__(self, facebook_helper: FacebookHelper, album_id: str,
                 video_title: str, reactions_threshold: int = 50,
                 time_threshold: timedelta = timedelta(days=1),
                 yet_to_check_file: str = "bofc.json",
                 store_best_ofs: bool = True, working_dir: Path = None):
        """
        Constructor
        :param facebook_helper: Helper to gather data and post it to Facebook
        :param album_id: Facebook album id where to repost best of frames
        :param video_title: Move/video title for the frames
        :param reactions_threshold: Threshold determining which frames should be reposted
        :param time_threshold: Time after which a frame's reactions count can be compared with the threshold and
        reposted
        :param yet_to_check_file: file where the queued frames data will be stored for later restarts
        :param store_best_ofs: determines if the best of frames should also be stored locally for later use
        """
        super().__init__(working_dir=working_dir)
        self.facebook_helper: FacebookHelper = facebook_helper
        self.album_id: str = album_id
        self.video_title: str = video_title
        self.reactions_threshold: int = reactions_threshold
        self.time_threshold: timedelta = time_threshold
        normalized_video_title = slugify.slugify(f"Best of {self.video_title}")
        self.working_dir = self.working_dir.joinpath(normalized_video_title)
        self.yet_to_check_file: Path = self.working_dir.joinpath(yet_to_check_file)
        self.yet_to_check: List[FacebookFrame] = []
        self.album_path: Path = self.working_dir.joinpath("album")
        self.frames_dir: Path = self.working_dir.joinpath("frames_to_check")
        self.store_best_ofs: bool = store_best_ofs

        self.logger.info(f"Best of reposting is enabled with a threshold of {self.reactions_threshold} reactions and "
                         f"a time threshold of {self.time_threshold}.")
        self.logger.info(f"Best ofs will be saved locally in the directory '{self.album_path}' and "
                         f"reuploaded in the album with id {self.album_id}.")
        os.makedirs(self.working_dir, exist_ok=True)
        os.makedirs(self.album_path, exist_ok=True)
        os.makedirs(self.frames_dir, exist_ok=True)

    def _check_for_existing_status(self) -> None:
        """
        Checks if a status file already exists in the local file system, and loads it if so.
        """
        if os.path.exists(self.yet_to_check_file):
            self.logger.info(f"Found existing {self.yet_to_check_file} file for best of checks, "
                             f"trying to load it...")
            self.yet_to_check: List[FacebookFrame] = utils.load_obj_from_json_file(self.yet_to_check_file)
            self.yet_to_check.sort(key=lambda yet_to_check_frame: yet_to_check_frame.post_time)

    def _advance_bests(self) -> None:
        """
        Checks if there are frames to repost into the best-of album, and posts them if so.
        """
        self.logger.info(f"Checking for best of reuploading...")
        try:
            while len(self.yet_to_check) > 0 and (self._check_and_post(self.yet_to_check[0])):
                self.yet_to_check.pop(0)
                utils.safe_json_dump(self.yet_to_check_file, self.yet_to_check)
        except FacebookError:
            self.logger.warning("There was a problem during the check of best-ofs", exc_info=True)
        self.logger.info("Done checking for best-ofs.")

    def _check_and_post(self, frame: FacebookFrame) -> bool:
        """
        Checks if the time threshold time has passed, then checks the reactions count against the threshold,
        and if so, reuploads the frame in the best-of album
        :param frame: the frame to be checked
        :return: True if the frame has been checked, False if it's still too early
        """
        elapsed_time = datetime.now() - frame.post_time
        if elapsed_time < self.time_threshold:
            return False
        self.logger.info(f"Checking entry {frame}...")
        if os.path.exists(frame.local_file):
            reactions_total = self.facebook_helper.get_reactions_total_count(frame.post_id)
            if reactions_total > self.reactions_threshold:
                self.logger.info(f"Uploading frame {frame.local_file} to best of album...")
                message = f"Reactions after {int(elapsed_time.total_seconds() / 3600)} hours : " \
                          f"{reactions_total}.\n" + \
                          f"Original post: {frame.url}\n\n" + \
                          frame.text
                self.facebook_helper.post_photo(frame.local_file, message, self.album_id)
                shutil.copyfile(frame.local_file,
                                os.path.join(self.album_path,
                                             f"Frame {frame.number} "
                                             f"id {frame.post_id} "
                                             f"reactions {reactions_total}"))
            os.remove(frame.local_file)
        else:
            self.logger.warning(f"File {frame.local_file} is missing. Skipping best of check...")
        return True

    def _queue_frame_for_check(self, frame: FacebookFrame) -> None:
        """
        Adds a frame data to the queue for later checking
        :param frame: the frame to be queued
        """
        self.logger.info(f"Queueing frame {frame.number} for best of checking...")
        # make shallow copy so no parameters from original are overwritten
        frame = copy.copy(frame)
        # copy frame to temp dir
        new_file_path = self.frames_dir.joinpath(frame.local_file.name).resolve(strict=False)
        shutil.copyfile(frame.local_file, new_file_path)
        frame.local_file = new_file_path
        self.yet_to_check.append(frame)
        utils.safe_json_dump(self.yet_to_check_file, self.yet_to_check)

    def _handle_quicker(self) -> None:
        """
        Halves the time threshold and starts a loop to check the remaining frames. Used after the framebot has
        finished posting.
        """
        self.time_threshold /= 2
        while len(self.yet_to_check) > 0:
            self._advance_bests()
            if len(self.yet_to_check) > 0:
                self.logger.info(
                    f"There are still {len(self.yet_to_check)} frames to check for best of reuploading. "
                    f"Sleeping for one hour...")
                time.sleep(timedelta(hours=1).total_seconds())

    def before_upload_loop(self) -> None:
        self._check_for_existing_status()

    def before_frame_upload(self, frame: FacebookFrame) -> None:
        self._advance_bests()

    def after_frame_upload(self, frame: FacebookFrame) -> None:
        self._queue_frame_for_check(frame)

    def after_upload_loop(self) -> None:
        self._handle_quicker()


class MirroredFramePoster(FrameBotPlugin):
    """
    Randomly mirrors and posts a frame with a random chance.
    """

    def __init__(self, facebook_helper: FacebookHelper, album_id: str, ratio: float = 0.5,
                 bot_name: str = "MirrorBot", mirror_original_message: bool = True,
                 extra_message: str = None):
        """
        Constructor
        :param facebook_helper: Helper to gather data and post it to Facebook
        :param album_id: The Facebook album where to post mirrored photos
        :param ratio: The percentage a frame has to be reposted mirrored
        :param bot_name: The bot's name. Used in the default extra message
        :param mirror_original_message: Also mirrors the original text message along with the image
        :param extra_message: Message to attach to the frame. Default adds the bot's name as a sort of signature
        """
        super().__init__()
        self.facebook_helper: FacebookHelper = facebook_helper
        self.album_id: str = album_id
        self.ratio: float = ratio
        self.bot_name: str = bot_name
        self.mirror_original_message: bool = mirror_original_message
        if extra_message is None:
            extra_message = f"Just a randomly mirrored image.\n-{self.bot_name}"
        self.extra_message = extra_message
        self.logger.info(f"Random mirroring is enabled with ratio {self.ratio}. Mirrored frames will be "
                         f"posted to the album with id {self.album_id}.")

    def _mirror_frame(self, frame: FacebookFrame) -> Image:
        """
        Mirrors a frame and returns it
        :param frame: Frame to be mirrored
        :return the mirrored frame
        """
        im = Image.open(frame.local_file)
        flipped_half = ImageOps.mirror(im.crop((0, 0, im.size[0] // 2, im.size[1])))
        im.paste(flipped_half, (im.size[0] // 2, 0))
        return im

    def _generate_message(self, frame: FacebookFrame) -> str:
        """
        Generates a message from the mirrored frame, depending on the bot settings
        :param frame: the frame to be mirrored
        :return: the generated message
        """
        lines = frame.text.split("\n")
        message = ""
        if self.mirror_original_message:
            message += "\n".join([line[:len(line) // 2] + line[len(line) // 2::-1] for line in lines])
        if self.extra_message != "":
            if self.mirror_original_message:
                message += "\n\n"
            message += self.extra_message
        return message

    def after_frame_upload(self, frame: FacebookFrame) -> None:
        if random() >= (1 - self.ratio / 100):
            self.logger.info("Posting mirrored frame...")
            mirrored_frame = self._mirror_frame(frame)
            frame_text = self._generate_message(frame)
            self.facebook_helper.post_photo(mirrored_frame, frame_text, self.album_id)


class AlternateFrameCommentPoster(FrameBotPlugin):
    """
    Plugin posting as a comment an alternate version of the same frame, taken from a given directory.
    The files in the alternate directory need to be named as the ones in the main frames directory.
    """

    def __init__(self, facebook_helper: FacebookHelper, alternate_frames_directory: Path, delete_files: bool = False,
                 message_generator: Union[str, Callable[[FacebookFrame], Union[str, None]]] = lambda frame: frame.text):
        """
        Constructor
        :param facebook_helper: facebook_helper: Helper to gather data and post it to Facebook
        :param alternate_frames_directory: the directory where the alternate frames are picked from
        :param delete_files: determines if the alternate frame files need to be deleted after the comment is posted
        :param message_generator: function used to generate the comment message, starting from the input frame, or
        static string to use as comment message
        """
        super().__init__()
        if not alternate_frames_directory.exists():
            raise ValueError(f"Can't initialize AlternateFrameCommentPoster plugin. Provided path for alternate frames"
                             f" doesn't exist: {alternate_frames_directory}")
        self.helper = facebook_helper
        self.frames_directory = alternate_frames_directory
        self.delete_files = delete_files
        self.message_generator = (lambda frame: message_generator) if type(message_generator) is str else \
            message_generator
        self.logger.info(f"Alternate frame posting is enabled. Will post alternate frames as post comments from the"
                         f" directory {self.frames_directory}. Alternate frames will"
                         f" {'' if self.delete_files else 'not'} be deleted.")

    def after_frame_upload(self, frame: FacebookFrame) -> None:
        alternate_frame_path = self.frames_directory.joinpath(frame.local_file.name)
        self.logger.info(f"Posting alternate frame from location {alternate_frame_path}...")
        if not alternate_frame_path.exists():
            raise FileNotFoundError(f"Alternate frame not found in path {alternate_frame_path}. Please check your"
                                    f" configuration and/or directories.")
        message = self.message_generator(frame)
        self.helper.post_comment(object_id=frame.photo_id, message=message, image=alternate_frame_path)
        if self.delete_files:
            alternate_frame_path.unlink()



import unittest
from datetime import timedelta
from io import BytesIO, BufferedReader, IOBase
from pathlib import Path
from typing import Union
from unittest import TestCase
from unittest.mock import MagicMock, patch, DEFAULT, call

from PIL import Image
from pyfacebook import FacebookError

from framebot.social import FacebookHelper, open_image_stream
from test import RESOURCES_DIR

PHOTO_ID = "1"
POST_ID = "123_" + PHOTO_ID
REACTIONS = 1230

ACCESS_TOKEN = "dummy"
PAGE_ID = "dummy_id"

DUMMY_IMAGE = RESOURCES_DIR.joinpath("dummy.jpg")


class TestFacebookHelper(TestCase):

    @patch("framebot.social.GraphAPI")
    def setUp(self, mock_graph_api_class: MagicMock) -> None:
        self.testee = FacebookHelper(access_token=ACCESS_TOKEN, page_id=PAGE_ID)
        self.mock_graph: MagicMock = self.testee.graph

    def test_post_photo(self):
        self._test_post_photo(DUMMY_IMAGE)
        self._test_post_photo(DUMMY_IMAGE, "some_album")

    def _test_post_photo(self, src_image: Path, album_id=None):
        mock_method: MagicMock = self.mock_graph.post_object
        mock_method.return_value = {"id": PHOTO_ID, "post_id": POST_ID}

        test_message = "Test photo upload"
        response = self.testee.post_photo(src_image, test_message, album_id)
        self.assertEqual(response.photo_id, PHOTO_ID)
        self.assertEqual(response.post_id, POST_ID)

        mock_method.assert_called_once()
        call_args = mock_method.call_args
        if type(call_args.kwargs) == dict:
            call_args = call_args.kwargs
        else:
            # python 3.7
            call_args = call_args[1]
        image = call_args['files']['source']

        self.assertTrue(image.closed)
        self.assertEqual(str(src_image), image.name)

        self.assertEqual(test_message, call_args['data']['message'])
        self.assertEqual((PAGE_ID if album_id is None else album_id), call_args['object_id'])
        self.assertEqual("photos", call_args['connection'])

        mock_method.reset_mock()

    def test_post_with_retry(self):
        object_id = "id"
        connection = "connection"
        files = {"source": "fake"}
        data = {"message": "test message"}

        mock_method: MagicMock = self.mock_graph.post_object
        mock_method.side_effect = FileNotFoundError

        self.assertRaises(FileNotFoundError, self.testee._post_with_retry, object_id=object_id,
                          connection=connection, files=files, data=data)
        mock_method.assert_called_once_with(object_id=object_id, connection=connection, files=files, data=data)

        mock_method.reset_mock()

        fake_error = FacebookError(kwargs={
            "error": {
                "code": 1,
                "message": "Not really an error"
            }
        })

        mock_method.side_effect = fake_error
        max_retries = 2
        retry_time = timedelta()
        self.assertRaises(FacebookError, self.testee._post_with_retry, object_id=object_id, connection=connection,
                          max_retries=max_retries, retry_time=retry_time)
        self.assertEqual(max_retries + 1, mock_method.call_count)

        mock_method.reset_mock()
        expected_result = {"id": object_id}
        mock_method.side_effect = [fake_error, expected_result]
        result = self.testee._post_with_retry(object_id=object_id, connection=connection, max_retries=max_retries,
                                              retry_time=retry_time)
        self.assertEqual(max_retries, mock_method.call_count)
        self.assertEqual(expected_result, result)
        mock_method.assert_called_with(object_id=object_id, connection=connection, files=None, data=None)

        # access token expired
        mock_method.reset_mock()
        fake_error.code = 190
        mock_method.side_effect = fake_error
        self.assertRaises(FacebookError, self.testee._post_with_retry, object_id, connection)

    def test_get_reactions_count(self):
        mock_method: MagicMock = self.mock_graph.get_object
        mock_method.return_value = {"reactions": {"summary": {"total_count": REACTIONS}}}

        reactions = self.testee.get_reactions_total_count(POST_ID)

        self.assertEqual(REACTIONS, reactions)
        mock_method.assert_called_once_with(object_id=POST_ID, fields="reactions.summary(total_count)")

        mock_method.reset_mock()
        # called with a photo id
        story_id = "story_id"
        mock_method.side_effect = [FacebookError(kwargs={
            "error": {
                "code": 100,
                "message": "Tried accessing nonexisting field (reactions)"
            }
        }),
            {"page_story_id": story_id},
            {"reactions": {"summary": {"total_count": REACTIONS}}}
        ]
        reactions = self.testee.get_reactions_total_count(POST_ID)
        self.assertEqual(REACTIONS, reactions)
        self.assertEqual(3, mock_method.call_count)
        expected_calls = [
            call(object_id=POST_ID, fields="reactions.summary(total_count)"),
            call(object_id=POST_ID, fields="page_story_id"),
            call(object_id=story_id, fields="reactions.summary(total_count)")
        ]
        mock_method.assert_has_calls(expected_calls)

    def test_post_comment(self):
        # Image and message both None
        post_id = "post_id"
        self.assertRaises(ValueError, self.testee.post_comment, post_id)

        mock_method: MagicMock = self.mock_graph.post_object
        expected_comment_id = "comment_id"
        mock_method.return_value = {"id": expected_comment_id}

        # only image
        comment_id = self.testee.post_comment(object_id=post_id, image=DUMMY_IMAGE)
        self.assertEqual(expected_comment_id, comment_id)
        mock_method.assert_called_once()

        # only message
        comment_id = self.testee.post_comment(object_id=post_id, message="message")
        self.assertEqual(expected_comment_id, comment_id)
        self.assertEqual(2, mock_method.call_count)

        # both image and message
        comment_id = self.testee.post_comment(object_id=post_id, image=DUMMY_IMAGE, message="message")
        self.assertEqual(expected_comment_id, comment_id)
        self.assertEqual(3, mock_method.call_count)

        # check empty string conversion
        comment_id = self.testee.post_comment(object_id=post_id, image=DUMMY_IMAGE, message="")
        self.assertEqual(expected_comment_id, comment_id)
        self.assertEqual(4, mock_method.call_count)
        call_args = mock_method.call_args
        if type(call_args.kwargs) == dict:
            call_args = call_args.kwargs
        else:
            # python 3.7
            call_args = call_args[1]
        self.assertEqual({"message": None}, call_args["data"])



class TestStaticMethods(TestCase):

    def test_open_image_stream(self):
        for image_variant in [DUMMY_IMAGE, str(DUMMY_IMAGE), Image.open(DUMMY_IMAGE)]:
            with open_image_stream(image_variant) as im_stream:
                self.assertTrue(issubclass(type(im_stream), IOBase))
            self.assertTrue(im_stream.closed)


if __name__ == '__main__':
    unittest.main()



"""
Contains data classes definitions
"""
from __future__ import annotations

from datetime import datetime
from pathlib import Path
from typing import Union, Optional


class FacebookFrame(object):
    """
    Object representing a Facebook frame (image)
    """

    def __init__(self, number: int, local_file: Union[str | Path]):
        """
        Constructor
        :param number: the frame number
        :param local_file: the local file where the image is stored
        """
        self.number: int = number
        self.local_file: Path = local_file if type(local_file) is Path else Path(local_file)
        self.photo_id: Optional[str] = None
        self.post_id: Optional[str] = None
        self.url: Optional[str] = None
        self.text: Optional[str] = None
        self.post_time: Optional[datetime] = None

    def __eq__(self, o: object) -> bool:
        if type(o) is type(self):
            return self.__dict__ == o.__dict__
        return NotImplemented

    def __str__(self) -> str:
        return str({
            "number": self.number,
            "local_file": self.local_file,
            "text": self.text,
            "post_id": self.photo_id,
            "post_time": self.post_time,
            "story_id": self.photo_id
        })



import datetime
import inspect
import shutil
import unittest
import os
from datetime import timedelta, datetime
from pathlib import Path
from typing import Callable
from unittest import TestCase
from unittest.mock import Mock, patch, DEFAULT
from pyfacebook import FacebookError

from PIL import Image, ImageOps
from framebot import utils

from framebot.model import FacebookFrame
from framebot.plugins import BestOfReposter, MirroredFramePoster, FileWritingFrameBotPlugin, AlternateFrameCommentPoster
from framebot.social import FacebookHelper
from test import RESOURCES_DIR
from test.utils_for_tests import FileWritingTestCase, generate_test_frame


class TestFileWritingFrameBotPlugin(FileWritingTestCase):

    def test_init(self):
        testee = FileWritingFrameBotPlugin(working_dir=self.test_dir)
        self.assertTrue(testee.working_dir.exists())


class TestBestOfReposter(FileWritingTestCase):

    def setUp(self) -> None:
        super().setUp()
        self.facebook_helper = Mock(spec=FacebookHelper)
        self.video_title = "test"
        self.test_bofc_path = RESOURCES_DIR.joinpath("plugins").joinpath("best_of_reposter").joinpath("bofc.json")
        self.test_frames = utils.load_obj_from_json_file(self.test_bofc_path)
        self.testee = BestOfReposter(facebook_helper=self.facebook_helper, album_id="album_id",
                                     video_title=self.video_title, working_dir=self.test_dir)
        self.test_frame = generate_test_frame()
        self.test_frame.post_time = datetime.now()
        self.test_frame.url = "https://example.com"

    def test_defaults(self):
        default_reactions_threshold = 50
        default_time_threshold = timedelta(days=1)
        default_yet_to_check_file = "bofc.json"
        default_store_best_ofs = True

        self.assertTrue(self.testee.album_path.exists())
        self.assertTrue(self.testee.frames_dir.exists())

        self.assertEqual(default_reactions_threshold, self.testee.reactions_threshold)
        self.assertEqual(default_time_threshold, self.testee.time_threshold)
        self.assertEqual(self.testee.working_dir.joinpath(default_yet_to_check_file),
                         self.testee.yet_to_check_file.absolute())
        self.assertEqual(default_store_best_ofs, self.testee.store_best_ofs)

    def test_check_for_existing_status(self):
        # asserting no exceptions are raised
        self.testee._check_for_existing_status()

        # asserting correct loading
        shutil.copy(self.test_bofc_path, self.testee.yet_to_check_file)
        self.testee._check_for_existing_status()
        self.assertEqual(len(self.test_frames), len(self.testee.yet_to_check))
        for frame in self.testee.yet_to_check:
            self.assertEqual(FacebookFrame, type(frame))
        self.assertEqual(self.test_frames, self.testee.yet_to_check)
        self.assertTrue(all(self.testee.yet_to_check[i].post_time < self.testee.yet_to_check[i + 1].post_time
                            for i in range(len(self.testee.yet_to_check) - 1)))

    @patch("shutil.copyfile")
    @patch("os.remove")
    def test_check_and_post(self, mock_remove: Mock, mock_copyfile: Mock):
        # too early
        self.testee.time_threshold = timedelta(days=1)
        self.assertFalse(self.testee._check_and_post(self.test_frame))

        exists_method = "os.path.exists"
        # not enough reactions
        with patch(exists_method, return_value=True) as mock_exists:
            reactions = 99
            self.facebook_helper.get_reactions_total_count.return_value = reactions
            self.test_frame.post_time = datetime.now() - timedelta(days=2)
            self.testee.reactions_threshold = 100

            self.assertTrue(self.testee._check_and_post(self.test_frame))
            mock_remove.assert_called_once_with(self.test_frame.local_file)
            self.assertFalse(self.facebook_helper.post_photo.called)
            self.assertFalse(mock_copyfile.called)
            mock_exists.assert_called_once_with(self.test_frame.local_file)

            # best of eligible
            self.testee.reactions_threshold = 98
            mock_remove.reset_mock()
            mock_exists.reset_mock()

            self.assertTrue(self.testee._check_and_post(self.test_frame))
            mock_remove.assert_called_once_with(self.test_frame.local_file)
            self.assertTrue(self.facebook_helper.post_photo.called)
            mock_copyfile.assert_called_once_with(self.test_frame.local_file,
                                                  os.path.join(self.testee.album_path,
                                                               f"Frame {self.test_frame.number} "
                                                               f"id {self.test_frame.post_id} "
                                                               f"reactions {reactions}"))
            mock_exists.assert_called_once_with(self.test_frame.local_file)

        # missing file
        with patch(exists_method, return_value=False) as mock_exists:
            self.assertTrue(self.testee._check_and_post(self.test_frame))
            mock_exists.assert_called_once_with(self.test_frame.local_file)

    @patch("framebot.utils.safe_json_dump")
    def test_advance_bests(self, mock_json_dump: Mock):
        self.testee.yet_to_check = self.test_frames
        mock_check_and_post = Mock()
        self.testee._check_and_post = mock_check_and_post
        mock_check_and_post.side_effect = [True, True, False]

        self.testee._advance_bests()
        self.assertEqual(1, len(self.testee.yet_to_check))
        self.assertEqual(2, mock_json_dump.call_count)
        self.assertEqual(3, mock_check_and_post.call_count)

        mock_check_and_post.reset_mock()
        mock_json_dump.reset_mock()

        # FacebookError raised
        mock_check_and_post.side_effect = FacebookError(kwargs={"error": {
            "message": "Not important",
            "code": 0
        }})
        expected_len = len(self.testee.yet_to_check)

        self.testee._advance_bests()
        mock_json_dump.assert_not_called()
        mock_check_and_post.assert_called_once()
        self.assertEqual(expected_len, len(self.testee.yet_to_check))

        mock_check_and_post.reset_mock()
        mock_json_dump.reset_mock()

        # last element checked
        mock_check_and_post.return_value = True
        mock_check_and_post.side_effect = [DEFAULT]
        self.testee._advance_bests()
        self.assertEqual(0, len(self.testee.yet_to_check))
        self.assertEqual(1, mock_json_dump.call_count)
        self.assertEqual(1, mock_check_and_post.call_count)

        mock_check_and_post.reset_mock()
        mock_json_dump.reset_mock()

        # nothing to check
        self.testee.yet_to_check = []
        self.testee._advance_bests()
        mock_check_and_post.assert_not_called()
        mock_json_dump.assert_not_called()

    @patch("framebot.utils.safe_json_dump")
    @patch("shutil.copyfile")
    def test_queue_for_frame_check(self, mock_copyfile: Mock, mock_json_dump: Mock):
        self.testee._queue_frame_for_check(self.test_frame)
        self.assertTrue(1, len(self.testee.yet_to_check))
        queued_frame = self.testee.yet_to_check[0]
        self.assertNotEqual(self.test_frame, queued_frame)
        new_frame_path = self.testee.frames_dir.joinpath(self.test_frame.local_file.name)
        self.assertEqual(new_frame_path, queued_frame.local_file)
        mock_copyfile.assert_called_once_with(self.test_frame.local_file, new_frame_path)
        mock_json_dump.assert_called_once_with(self.testee.yet_to_check_file, self.testee.yet_to_check)

    @patch("time.sleep")
    def test_handle_quicker(self, mock_sleep: Mock):
        def mock_advance_bests_behavior():
            # first call
            if len(self.testee.yet_to_check) == 3:
                self.testee.yet_to_check.pop(0)
                self.testee.yet_to_check.pop(1)
            # second call
            else:
                self.testee.yet_to_check.pop(0)

        og_threshold = self.testee.time_threshold
        self.testee.yet_to_check = self.test_frames
        mock_advance_bests = Mock(side_effect=mock_advance_bests_behavior)
        self.testee._advance_bests = mock_advance_bests

        self.testee._handle_quicker()

        self.assertEqual(og_threshold // 2, self.testee.time_threshold)
        self.assertEqual(0, len(self.testee.yet_to_check))
        self.assertEqual(2, mock_advance_bests.call_count)
        self.assertEqual(1, mock_sleep.call_count)


class TestMirroredFramePoster(TestCase):

    def setUp(self) -> None:
        super(TestMirroredFramePoster, self).setUp()
        self.album_id = "id"
        self.mock_helper = Mock(spec=FacebookHelper)
        self.testee = MirroredFramePoster(album_id=self.album_id, facebook_helper=self.mock_helper)
        self.test_frame = generate_test_frame()

    def test_default_extra_message(self):
        default_bot_name = inspect.signature(self.testee.__class__.__init__).parameters['bot_name'].default
        self.assertEqual(f"Just a randomly mirrored image.\n-{default_bot_name}", self.testee.extra_message)

    def test_mirror_frame(self):
        test_frame_image = Image.open(self.test_frame.local_file)
        mirrored_frame_image = self.testee._mirror_frame(self.test_frame)
        self.assertEqual(test_frame_image.size, mirrored_frame_image.size)
        size = test_frame_image.size
        self.assertEqual(
            test_frame_image.crop((0, 0, size[0] // 2, size[1])),
            mirrored_frame_image.crop((0, 0, size[0] // 2, size[1]))
        )
        self.assertEqual(
            test_frame_image.crop((0, 0, size[0] // 2, size[1])),
            ImageOps.mirror(mirrored_frame_image.crop((size[0] // 2, 0, size[0], size[1])))
        )

    def test_generate_message(self):
        # default extra + mirror original
        self.testee.mirror_original_message = True
        lines = self.test_frame.text.split("\n")
        mirrored_original = "\n".join([line[:len(line) // 2] + line[len(line) // 2::-1] for line in lines])
        expected_message = f"{mirrored_original}\n\n{self.testee.extra_message}"
        self.assertEqual(expected_message, self.testee._generate_message(self.test_frame))

        # no original
        self.testee.mirror_original_message = False
        self.assertEqual(self.testee.extra_message, self.testee._generate_message(self.test_frame))

        # no message at all
        self.testee.extra_message = ""
        self.assertEqual("", self.testee._generate_message(self.test_frame))

        # only original
        self.testee.mirror_original_message = True
        self.assertEqual(mirrored_original, self.testee._generate_message(self.test_frame))

    def test_after_frame_upload(self):
        mock_mirror_frame = Mock(spec=self.testee._mirror_frame)
        mock_generate_message = Mock(spec=self.testee._generate_message)
        self.testee._mirror_frame = mock_mirror_frame
        self.testee._generate_message = mock_generate_message

        # don't post
        self.testee.ratio = 0
        self.testee.after_frame_upload(self.test_frame)
        mock_mirror_frame.assert_not_called()
        mock_generate_message.assert_not_called()

        # post
        self.testee.ratio = 100
        self.testee.after_frame_upload(self.test_frame)
        mock_mirror_frame.assert_called_once_with(self.test_frame)
        mock_generate_message.assert_called_once_with(self.test_frame)


class TestAlternateFrameCommentPoster(TestCase):

    def setUp(self) -> None:
        self.mock_helper = Mock(spec=FacebookHelper)
        self.alternate_directory: Mock = Mock(spec=Path)
        self.test_frame = generate_test_frame()

    def test_init(self):
        # test that generator generates singleton from static string
        generator_output = "static"
        testee = AlternateFrameCommentPoster(alternate_frames_directory=self.alternate_directory,
                                             facebook_helper=self.mock_helper, message_generator=generator_output)
        self.assertTrue(callable(testee.message_generator))
        self.assertEqual(generator_output, testee.message_generator(None))

        # test default
        generator_output = "static"
        testee = AlternateFrameCommentPoster(alternate_frames_directory=self.alternate_directory,
                                             facebook_helper=self.mock_helper)
        self.assertEqual(self.test_frame.text, testee.message_generator(self.test_frame))

    def test_after_frame_upload(self):
        testee = AlternateFrameCommentPoster(alternate_frames_directory=self.alternate_directory,
                                             facebook_helper=self.mock_helper)
        mock_alternate_path: Mock = Mock(spec=Path)
        self.alternate_directory.joinpath.return_value = mock_alternate_path

        # file not found
        mock_alternate_path.exists.return_value = False

        self.assertRaises(FileNotFoundError, testee.after_frame_upload, self.test_frame)

        # normal workflow, no delete
        mock_alternate_path.exists.return_value = True
        testee.delete_files = False
        testee.after_frame_upload(self.test_frame)
        self.mock_helper.post_comment.assert_called_once_with(
            object_id=self.test_frame.photo_id, message=testee.message_generator(self.test_frame),
            image=mock_alternate_path)
        mock_alternate_path.unlink.assert_not_called()

        # normal workflow, delete
        testee.delete_files = True
        testee.after_frame_upload(self.test_frame)
        self.assertEqual(2, self.mock_helper.post_comment.call_count)
        mock_alternate_path.unlink.assert_called_once()


if __name__ == '__main__':
    unittest.main()



import datetime
import json
import os
import shutil
from configparser import ConfigParser
from pathlib import Path
from typing import Union
from unittest.mock import Mock

from slugify import slugify

from ..model import FacebookFrame
from ..plugins import BestOfReposter
from ..utils import safe_json_dump
from ..framebots import LAST_FRAME_UPLOADED_FILE, SimpleFrameBot
from ..social import FacebookHelper


def _get_old_album_name(movie_title: str) -> str:
    return "".join(
        x for x in f"Bestof_{movie_title.replace(os.path.sep, '-').replace(' ', '_')}" if
        (x.isalnum()) or x in "._- ")


def backup_bof_reposter_stuff(source_dir: Union[Path, str]) -> None:
    target_backup_dir = Path(source_dir).joinpath("migration_backup")
    target_backup_dir.mkdir()
    shutil.copy(source_dir.joinpath("bofc.json"), target_backup_dir)
    shutil.copytree(source_dir.joinpath("albums"), target_backup_dir.joinpath("albums"))


def migrate_bofc_json(source_dir: Union[Path, str], target_dir: Union[Path, str], framebot: SimpleFrameBot) -> None:
    with open(source_dir.joinpath("bofc.json"), "r") as f:
        old_bofc = json.load(f)

    old_frames = old_bofc["list"]
    new_frames = []
    frames_to_check_path = target_dir.joinpath("frames_to_check")
    frames_to_check_path.mkdir()
    for frame in old_frames:
        frame_file_name = Path(frame["path"]).name
        new_frame_path = frames_to_check_path.joinpath(frame_file_name).resolve(strict=False)
        shutil.copy(source_dir.joinpath("frames").joinpath(frame_file_name), new_frame_path)
        new_frame = FacebookFrame(
            local_file=new_frame_path,
            number=frame["frame_number"]
        )
        new_frame.photo_id = frame["post_id"]
        new_frame.post_id = frame["post_id"]  # only valid during migration
        new_frame.url = f"https://facebook.com/{new_frame.photo_id}"
        new_frame.reactions_total = None
        new_frame.post_time = datetime.datetime.fromisoformat(frame["time"])
        new_frame.text = framebot._get_default_message(new_frame.number)
        new_frames.append(new_frame)
    safe_json_dump(target_dir.joinpath("bofc.json"), new_frames)


def _get_framebot(source_dir: Union[Path, str], config: ConfigParser) -> SimpleFrameBot:
    movie_title = config["bot_settings"]["movie_title"]
    frames_naming = config["bot_settings"]["frames_naming"]
    frames_ext = config["bot_settings"]["frames_ext"]
    framebot = SimpleFrameBot(video_title=movie_title, frames_naming=frames_naming, frames_ext=frames_ext,
                              facebook_helper=Mock(spec=FacebookHelper),
                              frames_directory=source_dir.joinpath("frames"),
                              working_dir=source_dir)
    assert framebot.total_frames_number > 0, "Framebot did not load the frames correctly!"
    return framebot


def migrate(source_dir: Union[Path, str], target_dir: Union[Path, str] = None):
    print(f"Migrating framebot configuration from {source_dir} to {target_dir}...")
    config_file = "config.ini"
    if target_dir is None:
        target_dir = source_dir
    source_dir = source_dir.absolute()
    target_dir = target_dir.absolute()
    if target_dir == source_dir:
        print("Backing up old files...")
        backup_bof_reposter_stuff(source_dir)
    else:
        old_frames_path = source_dir.joinpath("frames")
        new_frames_path = target_dir.joinpath("frames")

        print(f"Copying frames directory from {old_frames_path} to {new_frames_path}...")
        shutil.copytree(old_frames_path, new_frames_path)

        old_config_file = source_dir.joinpath(config_file)
        new_config_file = target_dir.joinpath(config_file)
        print(f"Copying config file from {old_config_file} to {new_config_file}...")
        shutil.copy(old_config_file, new_config_file)

        old_last_frame_uploaded_file = source_dir.joinpath(LAST_FRAME_UPLOADED_FILE)
        new_last_frame_uploaded_file = target_dir.joinpath(LAST_FRAME_UPLOADED_FILE)
        print(f"Copying config file from {old_last_frame_uploaded_file} to {new_last_frame_uploaded_file}...")
        shutil.copy(old_last_frame_uploaded_file, new_last_frame_uploaded_file)

    config = ConfigParser()
    config.read(source_dir.joinpath(config_file), encoding='utf-8')
    movie_title = config["bot_settings"]["movie_title"]
    framebot = _get_framebot(source_dir, config)

    old_album_path = source_dir.joinpath("albums").joinpath(_get_old_album_name(movie_title))
    plugins_dir = target_dir.joinpath("plugins")
    bof_reposter_dir = plugins_dir.joinpath(BestOfReposter.__name__).joinpath(slugify(f"Best of {movie_title}"))
    bof_reposter_dir.mkdir(parents=True)

    print("Migrating bofc.json file and yet to check frames...")
    migrate_bofc_json(source_dir, bof_reposter_dir, framebot)

    new_album_path = bof_reposter_dir.joinpath("album")
    print(f"Copying bof album directory from {old_album_path} to {new_album_path}...")
    shutil.copytree(old_album_path, new_album_path)

    print("Done.")



"""
Contains framebots implementations
"""
import datetime
import os
import re
import time
from datetime import timedelta, datetime
from pathlib import Path
from re import Pattern
from typing import List, Union

from . import utils, DEFAULT_WORKING_DIR
from .model import FacebookFrame
from .plugins import FrameBotPlugin
from .social import FacebookHelper

LAST_FRAME_UPLOADED_FILE = "last_frame_uploaded"


def _get_filename(full_path: Union[str, Path]):
    """

    :param full_path:
    :return:
    """
    if issubclass(type(full_path), Path):
        full_path = str(full_path)
    return full_path[full_path.rfind(os.path.sep) + 1:]


# placeholder
class Framebot(utils.LoggingObject):
    """
    Placeholder class for later extension
    """
    pass


class SimpleFrameBot(Framebot):
    """
    Uploads frames from a fixed directory
    """

    def __init__(self, facebook_helper: FacebookHelper, video_title: str, frames_directory: Union[str, Path] = "frames",
                 frames_ext: str = "jpg", frames_naming: str = "$N$",
                 upload_interval: timedelta = timedelta(seconds=150), bot_name: str = "Bot",
                 delete_files: bool = False, plugins: List[FrameBotPlugin] = None,
                 working_dir: Union[Path, str] = DEFAULT_WORKING_DIR):
        """"
        :param facebook_helper: Helper to gather data and post it to Facebook
        :param video_title: Title of the movie/episode/whatever you want to post. Will be showed in the posts
            description
        :param frames_directory: Directory where the frame files are stored
        :param frames_ext: Extension of the frame files
        :param frames_naming: Naming pattern of the frame files e.g. frame$N$
        :param upload_interval: time interval between one frame and the other
        :param bot_name: bot's name, currently used only in the mirrored posts
        :param delete_files: if this flag is enabled, the frame files will be deleted after those served their purpose
        :param plugins: plugins to extend the bot's functionalities
        """
        super().__init__()
        self.facebook_helper = facebook_helper
        # Bot base settings
        self.bot_name = bot_name
        self.video_title = video_title
        self.upload_interval = upload_interval
        self.delete_files = delete_files
        self.frames_directory = (frames_directory if type(frames_directory) is Path else Path(frames_directory))\
            .resolve(strict=True)
        self.frames_ext = frames_ext
        self.frames_naming = frames_naming
        self.working_dir = working_dir.resolve(strict=False)
        self.working_dir.mkdir(parents=True, exist_ok=True)
        self.last_frame_uploaded_file = self.working_dir.joinpath(LAST_FRAME_UPLOADED_FILE)
        if plugins is None:
            plugins = []
        self.plugins: List[FrameBotPlugin] = plugins
        self._init_status()
        self._init_frames()

        self._log_parameters()
        self.logger.info("Done initializing.")

    @property
    def frames_naming(self) -> Pattern:
        return self._frames_naming

    @frames_naming.setter
    def frames_naming(self, naming_str: str) -> None:
        if "$N$" not in naming_str:
            raise ValueError("Frames naming must contain the $N$ placeholder to determine where the frame number"
                             " should be extracted from.")
        naming_str = naming_str.replace("$N$", "(\\d+)")
        expr = f"^{naming_str}\\.{self.frames_ext}$"
        self._frames_naming = re.compile(expr)

    def _init_frames(self) -> None:
        """
        Initializes the frames list and total frames number
        """
        self.frames: List[FacebookFrame] = []
        for frame_path in self.frames_directory.glob(f"*.{self.frames_ext}"):
            try:
                index_number = self._get_frame_index_number(frame_path)
                if index_number > self.last_frame_uploaded:
                    self.frames.append(FacebookFrame(index_number, frame_path.resolve(strict=True)))
            except AttributeError:
                # doesn't match the regex: not a valid frame
                self.logger.warning(f"File {frame_path} doesn't match the naming regex {self._frames_naming}. Bot will"
                                    f" not load it as a frame.")
        self.frames.sort(key=lambda frame: frame.number)
        if len(self.frames) == 0:
            self.total_frames_number = 0
        else:
            self.total_frames_number = self.frames[-1].number
        self.logger.info(f"Found {len(self.frames)} yet to upload frames.")

    def _init_status(self) -> None:
        """
        Check if stored status exists and loads it
        """
        if self.last_frame_uploaded_file.exists():
            with open(self.last_frame_uploaded_file) as f:
                self.last_frame_uploaded = int(f.read())
                self.logger.info(f"Last frame uploaded is {self.last_frame_uploaded}.")
        else:
            self.last_frame_uploaded = -1
            self.logger.info(f"Starting the bot from the first frame.")

    def _log_parameters(self) -> None:
        """
        Logs the loaded bot parameters
        """
        self.logger.info(f"Starting bot '{self.bot_name}'.")
        self.logger.info(f"Movie/video title is '{self.video_title}'.")
        self.logger.info(f"A frame will be posted every {self.upload_interval} seconds.")
        self.logger.info(f"Frames will {'' if self.delete_files else 'not'} be deleted after posting.")
        self.logger.info(f"Frames will be picked from the directory '{self.frames_directory}'. Will search for "
                         f".{self.frames_ext} files with naming pattern {self.frames_naming}.")
        plugin_names = [type(plugin).__name__ for plugin in self.plugins]
        self.logger.info(f"The following plugins are loaded: {plugin_names}")

    def _get_frame_index_number(self, file_path: Union[Path, str]) -> int:
        """
        Using the frame naming regexp, extracts the frame number from the filename
        :param file_path: The path of the frame file
        :return: The frame number
        """
        return int(self.frames_naming.search(_get_filename(file_path)).group(1))

    def _get_default_message(self, frame_number: int) -> str:
        """
        Generates the default message for a post
        :param frame_number: the index number of the frame
        :return: the post message
        """
        return f"{self.video_title}\nFrame {frame_number} of {self.total_frames_number}"

    def _update_last_frame_uploaded(self, number: int) -> None:
        """
        Stores the last uploaded frame number into a file for later resuming
        :param number: the latest frame uploaded's index number
        """
        with open(self.last_frame_uploaded_file, "w") as f:
            self.last_frame_uploaded = number
            f.write(str(number))

    def start(self) -> None:
        """
        Starts the framebot.
        """
        for plugin in self.plugins:
            plugin.before_upload_loop()
        self._upload_loop()
        for plugin in self.plugins:
            plugin.after_upload_loop()
        self.last_frame_uploaded_file.unlink()

    def _upload_loop(self) -> None:
        """
        The frame upload loop
        """
        self.logger.info("Starting upload loop.")
        for frame in self.frames:
            for plugin in self.plugins:
                plugin.before_frame_upload(frame)
            self._upload_frame(frame)
            for plugin in self.plugins:
                plugin.after_frame_upload(frame)
            if self.delete_files:
                os.remove(frame.local_file)
            if not (frame.number == self.total_frames_number):
                adjusted_pause = self._determine_adjusted_pause(frame)
                self.logger.info(f"Uploaded. Waiting {adjusted_pause} seconds before the next one...")
                time.sleep(adjusted_pause.total_seconds())
        self.logger.info("Upload loop over.")

    def _determine_adjusted_pause(self, last_posted_frame: FacebookFrame) -> timedelta:
        """
        Determines how long the bot should wait before posting the next frame, in order to be more regular
        with the expected post interval.
        :param last_posted_frame: the last posted frame
        :return: the adjusted pause
        """
        wanted_post_time = last_posted_frame.post_time + self.upload_interval
        now = datetime.now()
        if wanted_post_time <= now:
            return timedelta(seconds=0)
        return wanted_post_time - now

    def _upload_frame(self, frame: FacebookFrame) -> None:
        """
        Uploads a single frame
        :param frame: the frame to be uploaded
        """
        self.logger.info(f"Uploading frame {frame.number} of {self.total_frames_number}...")

        frame.text = self._get_default_message(frame.number)
        response = self.facebook_helper.post_photo(frame.local_file, frame.text)

        frame.photo_id = response.photo_id
        frame.post_id = response.post_id
        frame.url = f"https://facebook.com/{frame.photo_id}"
        frame.post_time = datetime.now()

        self._update_last_frame_uploaded(frame.number)



"""
Contains helper methods and classes for social media upload and data gathering
"""
import time
from contextlib import contextmanager
from datetime import timedelta
from io import BytesIO
from pathlib import Path
from typing import Union, Dict

from pyfacebook import GraphAPI, FacebookError
from PIL.Image import Image

from .utils import LoggingObject


class FacebookPostPhotoResponse:

    def __init__(self, photo_id: str, post_id: str):
        self.photo_id = photo_id
        self.post_id = post_id

    @classmethod
    def from_response_dict(cls, api_response: dict):
        return cls(api_response["id"], api_response["post_id"])

    def __eq__(self, o: object) -> bool:
        if type(o) is type(self):
            return self.__dict__ == o.__dict__
        return NotImplemented

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return repr(self.__dict__)


class FacebookHelper(LoggingObject):
    """
    Helper to interact with the Facebook Graph API
    """
    DEFAULT_RETRY_MINUTES = timedelta(minutes=1)
    DEFAULT_MAX_RETRIES = 5

    def __init__(self, access_token: str, page_id: str, timeout: timedelta = timedelta(seconds=20)):
        """
        Constructor
        :param access_token: Access token for the Facebook page
        :param page_id: Id of the Facebook page
        :param timeout: timeout for the http requests
        """
        super().__init__()
        self.access_token: str = access_token
        self.page_id: str = page_id
        # Ignoring the float warning here. The pyfacebook library incorrectly expects an integer, but then passes
        # the value to the requests module, which in turn expects a float
        self.graph: GraphAPI = GraphAPI(access_token=access_token, timeout=timeout.total_seconds())

        self.logger.info(f"Initialized GraphAPI for Facebook. Page id is {self.page_id}.")

    def post_photo(self, image: Union[Path, str, Image], message: str, album_id: str = None,
                   max_retries: int = DEFAULT_MAX_RETRIES, retry_time: timedelta = DEFAULT_RETRY_MINUTES) \
            -> FacebookPostPhotoResponse:
        """
        Uploads a photo to a specific album, or to the news feed if no album id is specified
        :param retry_time: time to wait if a failure occurs, before the next retry
        :param max_retries: max number of retries before giving up
        :param image: The image to be posted. Could be a path to an image file or a PIL Image
        :param message: The message used as image description
        :param album_id: The album where to post the image
        :return the response object containing photo id and post id
        """
        if album_id is None:
            album_id = self.page_id
        with open_image_stream(image) as im:
            response = self._post_with_retry(object_id=album_id, connection="photos", files={"source": im},
                                             data={"message": message}, max_retries=max_retries,
                                             retry_time=retry_time)
        return FacebookPostPhotoResponse.from_response_dict(response)

    def post_comment(self, object_id: str, image: Union[Path, str, Image] = None, message: str = None,
                     max_retries: int = DEFAULT_MAX_RETRIES, retry_time: timedelta = DEFAULT_RETRY_MINUTES) -> str:
        """
        Uploads a comment to a specific post. At least one between image and message must not be None.
        :param retry_time: time to wait if a failure occurs, before the next retry
        :param max_retries: max number of retries before giving up
        :param image: The image to be posted. Could be a path to an image file or a PIL Image, or None
        :param message: The comment message, if any
        :param object_id: The post where to append the comment
        :return the comment id
        """
        if message == "":
            # empty string causes an exception with the fb api
            message = None
        if image is None and message is None:
            raise ValueError("At least one between image and message must not be None.")
        if image is not None:
            with open_image_stream(image) as im:
                response = self._post_with_retry(object_id=object_id, connection="comments", files={"source": im},
                                                 data={"message": message}, max_retries=max_retries,
                                                 retry_time=retry_time)
        else:
            response = self._post_with_retry(object_id=object_id, connection="comments", data={"message": message},
                                             max_retries=max_retries, retry_time=retry_time)
        return response['id']

    def _post_with_retry(self, object_id: str, connection: str, files: Dict = None, data: Dict = None,
                         max_retries: int = DEFAULT_MAX_RETRIES, retry_time: timedelta = DEFAULT_RETRY_MINUTES) -> Dict:
        retry_count = 0
        while True:
            try:
                return self.graph.post_object(object_id=object_id, connection=connection, files=files, data=data)
            except FacebookError as e:
                if e.code == 190:
                    self.logger.error("Expired access token. Cannot post")
                    raise e
                self.logger.warning("Exception occurred during photo upload.", exc_info=True)
                if retry_count < max_retries:
                    retry_secs = retry_time.total_seconds() if "spam" not in str(e) else retry_time.total_seconds() * 10
                    self.logger.warning(f"Retrying photo upload after {retry_secs} seconds.")
                    time.sleep(retry_secs)
                else:
                    self.logger.error("Unable to post even after several retries. Check what's happening.")
                    raise e
                retry_count += 1

    def get_reactions_total_count(self, post_id: str) -> int:
        """
        Gathers the total reactions count for a post
        :param post_id: the post's story id
        :return: the total reaction count
        """
        try:
            return self.graph.get_object(object_id=post_id, fields="reactions.summary(total_count)")["reactions"][
                "summary"]["total_count"]
        except FacebookError as e:
            if e.code == 100:  # nonexisting field, e.g. that's a normal photo id from old version of the bot
                self.logger.warning(f"Tried calling 'get_reactions_total_count' with a photo id instead of a post id."
                                    f"This should only happen right after migrating from an old framebot version.")
                return self.get_reactions_total_count(self._get_story_id(post_id))
            raise e

    def _get_story_id(self, object_id: str) -> str:
        """
        Returns the story id for a given post. Here for compatibility reasons after code changes.
        :param object_id: the object id
        :return: the story id
        """
        return self.graph.get_object(object_id=object_id, fields="page_story_id")["page_story_id"]


@contextmanager
def open_image_stream(image: Union[Path, str, Image]) -> Union[bytes, BytesIO]:
    if issubclass(type(image), (str, Path)):
        im_stream = open(image, "rb")
        output = im_stream
    else:
        im_stream = BytesIO()
        image.save(im_stream, "jpeg")
        im_stream.seek(0)
        output = im_stream
    try:
        yield output
    finally:
        im_stream.close()
